\section{Overview}

This manual describes an experimental typechecker and compiler for the typed
$\lambda$-calculus \FOMEGASUB{}, an extension of Girard's system
 \FOMEGA{}~\cite{GIRARD72} to include subtyping.  The ideas behind
this calculus are due to Cardelli, particularly to his 1988 paper,
``Structural Subtyping and the Notion of Power
Type''~\cite{Cardelli88a}; the extension of the subtype relation to
type operators was developed by Cardelli and
Mitchell~\cite{Cardelli90:Fomegasub,Mitchell90,BruceMitchell92}.
Cardelli~\cite{Cardelli90:Fomegasub} has given a more powerful
treatment of operator subtyping, including both monotonic and
antimonotonic subtyping in addition to pointwise subtyping.  The
version of \FOMEGASUB{} used here was obtained as an instance of the
definition of a ``Generalized Subtyping System''~\cite{Pierce92d}, a
generalized type system in the sense of
Barendregt~\cite{Barendregt:GTS} extended with Cardelli's powerkinds.

Further information on the calculus itself and its applications in
object-oriented programming can be found
in~\cite[etc.]{CardelliWegner85,Cardelli90:Fomegasub,CMMS91,CurienGhelli91,Ghelli90,GhelliPierce92,HofMannPierce92,Pierce92d,PierceTurner92,PierceTurner92b}.

\section{Copying}

This is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation -- either version 1, or (at your option) any later
version.

This software is distributed in the hope that they will be useful,
but without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose.  See
the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

\section{Interaction}

\subsection{The Read-Eval-Print Loop}

To run the compiler interactively, type <<fomega>>.  A welcome banner
will be printed, followed by a prompt.
At this point, the compiler is ready to accept expressions terminated
by semicolons: 
> plus 5 6;

> f = fun(i:Int) plus i 2;

> f 5;

\noindent Each expression in turn is parsed, typechecked, compiled, and
executed, and the result printed.  The top-level environment is then
extended with a binding of the given name (or ``<<it>>'' if no name was
given explicitly) to the resulting value.
> x = plus 5 6;

> plus x 1;

> plus it 2;

\noindent Types can be declared in the same way:

> P = Some(X) {| v:X, f:X->Int |};

> p1 = < Int,
>        {v=3, f=fun(i:Int) plus i 1}
>      >: P;

\noindent The comment character is \%.  Everything from the \% to the end of the
line is ignored.
> % The doubling function for integers:
> double = fun(i:Int) plus i i;

\noindent To exit the system, send an EOF (control-D) character.


\subsection{Commands}

\newcommand{\descline}[2]{{\tt #1} & \parbox[t]{4in}{#2}}
\renewcommand{\arraystretch}{1.8}

In addition to expressions and type definitions, the top-level
read-eval-print loop responds to commands of the following forms:

\begin{tabular}{ll}
\descline{use "file";}%
{Read the contents of {\tt file} as if they had
                        been typed directly to the system.  If {\tt file}
                        is not given with an extension, the suffix
                        ``{\tt .f}'' is added.  If ``file'' has no
                        explicit path and it 
                        cannot be found in the current directory, the
                        subdirectory {\tt examples} is also searched.
                        Nested uses of ``{\tt use}'' are allowed.
} \\
\descline{use file;}%
{Same as the above, where ``file'' contains no special
                characters.
} \\
\descline{!command [arg];}%
{Execute the internal command {\tt command} with the
optional argument {\tt arg}.   
(See Section~\ref{s:intcomm}.)}
\end{tabular}

\subsection{Command-Line Switches}

The following command-line switches are recognized by the compiler:

\begin{tabular}{ll}
\descline{-o file}%
{Redirect the typechecker's output to a file
                      named {\tt file}.
}
\\ 
\descline{-set flag}%
{Set one of the internal flags (see Section~\ref{s:intflags}).}
\\
\descline{-reset flag}%
{Reset an internal flag.}
\\
\descline{-latex}%
{Same as ``{\tt -set latex -set TERSE}''.}
\\
\descline{-sml}%
{Immediately escapes to the SML read-eval-print
                      loop instead of running the typechecker.  (This
                      is mainly useful when rebuilding the typechecker.)}
\end{tabular}

\medskip

\noindent
If the command line includes a filename argument
\begin{verbatim}
fomega stuff.f
\end{verbatim}%
then input is read from this file instead of from the terminal.


\subsection{Internal Flags and Commands}

The compiler provides a number of internal flags, mostly used to
control the display of internal diagnostics and debugging information.
Some useful ones are:\label{s:intflags}

\begin{tabular}{ll}
\descline{latex}{Controls latex interaction mode (see Section~\ref{s:latex}).}
\\    
\descline{TERSE}{Controls printing of banners and inessential diagnostics.}
\\
\descline{echo}{Controls the echoing of input lines when the {\tt latex}
flag is false.} 
\\
\descline{eval}{Controls the code generator.  When this flag is {\tt false},
expressions are typechecked but not evaluated.  This improves the
speed of compilation by a factor of three or more.} 
\end{tabular}

\medskip

\noindent
The useful internal commands include:\label{s:intcomm}

\begin{tabular}{ll}
\descline{besilent}{Temporarily suppress all output.  (This is
                        useful when typesetting latex documents, for
                        establishing bindings or setting flags that
                        should not appear in the final output.)}
\\
\descline{beloud}{Overrides a previous ``{\tt besilent}''.}
\\
\descline{reload}{Throws away all the identifiers, types, and type
abbreviations in the current environment --- including the pervasive 
identifiers --- and establishes a new pervasive environment by
loading the file {\tt perv.f} from the current directory.}
\\
\descline{export ["name"]}{Creates an executable image called {\tt name}
({\tt fomega} by default) in
the current directory, whose initial pervasive environment contains
all the bindings currently in effect.}
\\
\descline{oldsyntax}{Silences the warnings normally generated when the
                parser detects old-style syntactic constructions 
                (square brackets around type arguments, curly braces
                for records, etc.).}
\end{tabular}
    

\subsection{Latex Interaction Mode}\label{s:latex}

To support the typesetting of documents with lambda-calculus examples,
the compiler supports a ``latex mode,'' established using the
``<<-latex>>'' command-line switch or the ``<<set latex>>'' internal command.  

In latex mode, lines of input that do not begin with ``\verb'>'''
or ``\verb'>'\verb'>''' or ``\verb'>'\verb'<''' are passed straight
through from the input to the 
output.  The delimiters ``\verb'<'\verb'<''' and ``\verb'>'\verb'>''' can
be used to indicate expressions that are to be set in the same
typeface as the lambda-calculus portions of the input.

Lines beginning with ``\verb'>''' are echoed and then passed to the parser,
typechecker, and evaluator as if they had been typed without the \verb'>'.
The echoed input and the system's responses are enclosed in a <<verbatim>>
environment.

Lines beginning with ``\verb'>'\verb'>''' are passed to the parser
without echoing them. 

Lines beginning with ``\verb'>'\verb'<''' are echoed without passing
them to the parser.  This is useful in the rare case where the
resulting \TeX{} file should have a ``\verb'>' character in the first
column.

\subsection{Type Abbreviations and Type Hints}\label{s:hints}

Type abbreviations are a simple form of syntactic shorthand for
simplifying the task of writing down programs involving complicated
type expressions, which could otherwise get unmanageably large.

Once a type abbreviation has been introduced
> IntArrowInt = Int->Int;
expressions involving the abbreviation behave the same in all
contexts as expressions involving the original.  
Abbreviations are expanded as necessary when they are encountered
during type analysis.

The typechecker does go to some trouble not to expand abbreviations
unnecessarily, so that the type that is printed for an expression can
be given in terms that the programmer will find easy to understand:
> fun(f:IntArrowInt) {a=f, b=f};
But the typechecker does {\em not} attempt to recognize situations
where  ``collapsing'' of abbreviations is possible:
> fun(f:IntArrowInt) fun(i:Int) f (f i);

When too many type abbreviations get ``expanded away'' during
typechecking --- or when it would be more pleasant to have the type of an
expression printed in terms of an abbreviation instead of some longer
form that the synthesis algorithm discovers --- {\em type hints} can
be used.  An expression of the form
\begin{verbatim}
e :< T >
\end{verbatim}%
is analyzed by first synthesizing a type for <<e>> alone, then
checking that this type is a subtype of the given type <<T>> (raising
an exception if it is not), and finally returning <<T>> as the
synthesized type of <<e>>.  
> fun(f:IntArrowInt) 
>   (fun(i:Int) f (f i))
>   :<IntArrowInt>;
Note that the prececence of the \verb'<:...>' construct is weaker than
that of <<fun>>, so that the parentheses around the middle line are
required. 

\renewcommand{\arraystretch}{1}

\subsection{Printers}\label{s:printers}

The read-eval-print loop maintains a collection of ``printers'' that
it uses to display the results of evaluation.  Each printer is an
expression of type <<T->Top(*)>> for some <<T>>, called the ``trigger
type'' of the printer.  When an expression --- of type <<S>>, say ---
is entered by the user, the most recently defined printer whose
trigger type is a supertype of <<S>> is used.  After being evaluated,
the expression is passed to the printer, which is expected to display
it in some appropriate way by using the built-in <<write>> primitive
to send characters to the output stream.  

There are predefined printers for integers, booleans, colors, and
strings, and a trivial printer with trigger type <<Top(*)>> that is
used when no other printer is applicable.
> plus 5 3;
> concat "hello" "there";
> and true false;
> fun(i:Int) i;

For example, let's define the usual polymorphic type for the Church
numerals~\cite{Bohm85}:
> Nat = All(A) A -> (A->A) -> A;
> zeroN = fun(A) fun(z:A) fun(s:A->A) z;
> succN = fun(n:Nat) 
>          fun(A) fun(z:A) fun(s:A->A) 
>            n A (s z) s;
> two = succN (succN zeroN);
Now we can define a printing function for church numerals that applies a given
numeral to the integer zero and successor function:
> prNat = fun(n:Nat) write 
>                       (concat 
>                          (concat "[" 
>                                  (int_to_string (n Int 0 succ)))
>                          "]");
> prNat two;
(Exercise: where did the \verb'<val>' come from?)
If we install <<prNat>> as a printer, the results of calculations
involving church numerals will
henceforth be displayed in familiar arabic notation:
> printer prNat;
> two;
> succN (succN two);

\subsection{Escaping to SML}\label{s:prim}

When the \FOMEGASUB{} typechecker encounters an expression of the form 
\begin{verbatim}
><    << "stuff" : T >>
\end{verbatim}
it assumes that the type of the phrase is <<T>> with no further
analysis.  The code generator then simply inserts the string <<stuff>> into
the intermediate code for the expression, which is passed to the
Standard ML of New Jersey compiler for translation to native code.

Using this facility correctly requires detailed understanding of the
file ``<<eval.sml>> in the \FOMEGASUB{} compiler sources.  Using it
incorrectly will almost surely crash the running <<fomega>> image.
The file <<basics.f>>, which defines the basic pervasive types and
built-in identifiers, gives some examples.  Good luck.


