\begin{exercise}
Convince yourself that the first version we tried of the function for
``sending the  <<setX>> message to a point object'' 
\begin{verbatim}
   wrong'setX =
     fun(P<Object(PointM))
       fun(p:P)
         open p as <Rep,r> in
            fun(i:Int)
               <Rep, {state = r.methods.setX r.state i,
                      methods = r.methods}
              > : P
         end;
\end{verbatim}
is really ill typed according to the typing rules of \FOMEGASUB{}. 
\end{exercise}

\ifsolution
\noindent It may help to see how the <<fomega>> system responds to
this declaration:
>> !besilent;
> PointM = Fun(Rep) {| getX: Rep->Int, 
>                      setX: Rep->Int->Rep |};
>> !beloud;
>> reset haltonerror;
> wrong'setX =
>   fun(P<Object(PointM))
>     fun(p:P)
>       open p as <Rep,r> in
>          fun(i:Int)
>             <Rep, {state = r.methods.setX r.state i,
>                    methods = r.methods}
>            > : P
>       end;
>> set haltonerror;
\fi


\begin{exercise}
Extend your implementation of slideable points from
Exercise~\ref{ex:slidepoints} so that the functions <<SlidePoint'setX>>,
<<SlidePoint'getX>>, and <<SlidePoint'slide>> deal properly with
subtyping.  
\end{exercise}

>> SlidePointM = Fun(R) {|setX: R->Int->R, getX: R->Int, slide: R->Int->R|};

\ifsolution
> SlidePoint'setX =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>      fun(i:Int)
>        open p as <R,r> in
>          <R, 
>           {state = r.methods.setX r.state i,
>            methods = r.methods}
>          >: Object(M)
>       end;
> SlidePoint'slide =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>      fun(i:Int)
>        open p as <R,r> in
>          <R, 
>           {state = r.methods.slide r.state i,
>            methods = r.methods}
>          >: Object(M)
>      end;
> SlidePoint'getX =
>  fun(M<SlidePointM)
>    fun(p:Object(M))
>       open p as <R,r> in
>         r.methods.getX r.state
>       end;
\fi

\begin{exercise}
Give a type operator <<SlideCPointM>> capturing the interface
specification of slideable colored points.  Define the functions
<<SlideCPoint'setC>> and <<SlideCPoint'getC>> and build a typical
element of <<Object(SlideCPointM)>>.  Write a small program that uses
your functions to manipulate this object by sending it messages.
\end{exercise}

\ifsolution
> SlideCPointM = Fun(R) {|setX: R->Int->R, getX: R->Int, slide: R->Int->R,
>                         setC: R->Color->R, getC: R->Color|};

> SlideCPoint'setC =
>  fun(M<SlideCPointM)
>    fun(p:Object(M))
>      fun(c:Color)
>        open p as <R,r> in
>          <R, 
>           {state = r.methods.setC r.state c,
>            methods = r.methods}
>          >: Object(M)
>       end;

> SlideCPoint'getC =
>  fun(M<SlideCPointM)
>    fun(p:Object(M))
>       open p as <R,r> in
>         r.methods.getC r.state
>       end;

> c = < {|x:Int,color:Color|}, 
>       {state = {x=5,color=red},
>        methods = {setX = fun(s:{|x:Int,color:Color|}) fun(i:Int) 
>                           {x=i,color=s.color},
>                   getX = fun(s:{|x:Int,color:Color|}) 
>                           s.x,
>                   slide = fun(s:{|x:Int,color:Color|}) fun(i:Int) 
>                           {x=plus s.x i, color=blue},
>                   setC = fun(s:{|x:Int,color:Color|}) fun(c:Color) 
>                           {x=s.x,color=c},
>                   getC = fun(s:{|x:Int,color:Color|}) 
>                           s.color
>                  }}
>      > : Object(SlideCPointM);

> let c1 = SlidePoint'setX SlideCPointM c 3 in
> let c2 = SlideCPoint'setC SlideCPointM c1 yellow in
> let c3 = SlidePoint'slide SlideCPointM c2 3 in
>   SlideCPoint'getC SlideCPointM c3
> end end end;
Note that the <<slide>> method behaves the same as in the Smalltalk
example, but that this behavior is achieved ``directly'' in the
implementation of <<slide>>, not by inheritance.
\fi

\begin{exercise}
Note that we can also define the functions <<SlideCPoint'setX>>,
<<SlideCPoint'getX>>, and <<SlideCPoint'slide>>.  Does their behavior
differ from the corresponding methods for slideable points?  Is it
useful to have both sets of functions?  (If not, can you think of any
circumstances where it would be useful?)
\end{exercise}

\ifsolution
\noindent No, in situations where both
\begin{verbatim}
    SlidePoint'setX M p i
\end{verbatim}
and
\begin{verbatim}
    SlideCPoint'setX M p i
\end{verbatim}
are well typed, their behavior is exactly the same.  

It will always be the case that the different versions of a message
sending function will behave identically whenever their behavior can be
compared.  (In fact, this observation might be taken as part of the
{\em definition} of the operation of sending a message;
see~\cite{HofmannPierce92}.)  Nevertheless, it may still be useful to
have the different versions available, since the type returned by the
operation of sending a particular message may sometimes be refined by
subclasses.  For example, the <<getX>> method of instances of
<<SlideCPoint>> might be defined so that it returned a <<Real>> instead
of an <<Int>>.  
\fi


