functor Eval(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    structure TEnv: TENV
	    structure TypHelper: TYPHELPER
	    structure Trm: TRM
	    structure TrmHelper: TRMHELPER
	    structure Synth: SYNTH
	    sharing Typ.Globals = Globals
	        and TypHelper.Typ = Typ
	        and TypHelper.TEnv = TEnv
	        and TEnv.Typ = Typ
		and Trm.TypHelper = TypHelper
	        and Globals.Id = Typ.Id
	        and Typ.Id = TEnv.Id
	        and TEnv.Id = TypHelper.Id
	        and TrmHelper.Trm = Trm
	        and Synth.Trm = Trm
	        and Synth.TEnv = TEnv
	    ) : EVAL
	    = struct

structure Globals = Globals
structure Typ = Typ
structure TypHelper = TypHelper
structure TEnv = TEnv
structure Trm = Trm
structure TrmHelper = TrmHelper

open TrmHelper
open Trm

val nameprefix = "f'"

type code = string

datatype T = DEFN of Id.T * code * T
           | BOUND of Id.T * T
	   | NULL  of Id.T * T
	   | EMPTY 

type env = T

val empty_env = EMPTY

fun push_env_binding env i e = DEFN(i,e,env)

fun push_env_const env i = BOUND(i,env)

fun push_env_null env i = NULL(i,env)

exception VariableOutOfRange of int
exception NotATermVariable of int

fun case_binding env n fabbrev fbound = let
  fun l (EMPTY) i = raise VariableOutOfRange(i)
    | l (DEFN(_,e,env)) i = 
      if i = 0
        then fabbrev e
        else l env (i-1)
    | l (BOUND(_,env)) i = 
      if i = 0
        then fbound ()
        else l env (i-1)
    | l (NULL(_,env)) i = 
      if i = 0
        then raise NotATermVariable(i)
        else l env (i-1)
  in l env n
  end

exception Eval of unit->unit

fun find_field [] l = 
      raise Eval(fn()=> write ("Nonexistent field: " ^ (Id.tostr l)))
  | find_field ((l',e)::rest) l = 
      if Id.== l l'
         then e
         else find_field rest l

val DEBUG = createflag ["EVAL"] false
val DEBUGMUCHO = createflag ["MOREEVAL"] false

(* ---------------------------------------------------------------------- *)
(* Use the running SML as our back-end *)

(* $$$ Perhaps we should just turn off low-level printing while the
   use_stream is doing its stuff, and leave it turned on the rest
   of the time? *)
fun doit c = let 
  val s = open_string c
  in if !DEBUG
       then write ("--> " ^ c ^ "\n")
       else ();
     System.Compile.use_stream s;
     close_in s
  end

(* This one probably works with every version of the compiler:
fun doit c = let 
  val f = open_out "use.tmp"
  in 
     output(f,c); output(f,"\n");
     close_out f;
     if !DEBUG
       then write ("--> " ^ c ^ "\n")
       else ();
     use "use.tmp"
  end
*)

val cast = "cast"

val initsml = 
    "val write = Globals.write;\n"
  ^ "val cast = System.Unsafe.cast;\n"
  ^ "val assoc = Globals.assoc (fn(x:string)=>fn(y:string)=>x=y);\n"
  ^ "val app = fn f => fn a => (cast ((cast f) a));\n"
  ^ "val pr = fn s:string => (output(std_err,s); flush_out std_err);\n"
  ^ "exception DummyCode;\n"

val _ = resetproc(
          fn()=> let val old_SMLout = !System.Print.out
                 in System.Print.out := open_out "/dev/null";
                    doit initsml;
                    System.Print.out := old_SMLout
                 end)          

(* ---------------------------------------------------------------------- *)
(* Call-by-value *)

fun emit s = "(" ^ cast ^ "(" 
           ^ (mapfold (fn s=>s) (fn s1 => fn s2 => s1^s2) "" s)
           ^ ") : string)"

fun gen_cbv te env exp = let
  fun ev (VAR(i)) = 
      case_binding env i 
        (fn e' => e')
        (fn () => Id.tostr (TEnv.lookup_name te i))
    | ev (ABS(n,t,e1)) = 
        emit ["fn ", Id.tostr n, " => ", 
              gen_cbv (TEnv.push_binding te n t) (push_env_const env n) e1]
    | ev (APP(e1,e2)) = 
        emit ["(cast ",gen_cbv te env e1, ") ", gen_cbv te env e2]
    | ev (TABS(x,t,e1)) = 
        emit ["fn()=> ", 
              gen_cbv (TEnv.push_bound te x t) (push_env_null env x) e1]
    | ev (TAPP(e1,t2)) = 
        emit ["(cast ",gen_cbv te env e1, ") ", "()"]
    | ev (RCDV(fields)) = let
        val f = mapappend_tuple 
                  (fn(l,e) => ["(\"", Id.tostr l, "\",", gen_cbv te env e,")"])
                  ", " 
                  fields
        in emit (["["] @ f @ ["]"])
        end
    | ev (PROJ (e,l)) =
        emit ["assoc (cast ", gen_cbv te env e, ") \"", Id.tostr l, "\""]
    | ev (PACK(e,_,_)) = gen_cbv te env e
    | ev (OPEN(a,x,e,b)) = let
        in emit ["let val ", Id.tostr x, " = ", gen_cbv te env e, " in ", 
                 gen_cbv (TEnv.push_binding 
                             (TEnv.push_bound te a Typ.dummytype) 
                             x Typ.dummytype)
                         (push_env_const 
                             (push_env_null env a) x)
                         b,
                 " end"]
        end
    | ev (ASSERT(e,t)) = 
        gen_cbv te env e
    | ev (ABBREV(a,t,e)) = 
        gen_cbv te env (subst_top_type t e) 
    | ev (LET(x,e,b)) = 
        emit ["let val ", Id.tostr x, " = ", gen_cbv te env e, " in ", 
                 gen_cbv (TEnv.push_binding te x Typ.dummytype)
                         (push_env_const env x)
                         b,
                 " end"]
    | ev (PRIM(p,t)) = 
        emit [p]
  in wrap DEBUGMUCHO "gen_cbv"
    (fn () => ev exp)
    (fn () => prt te exp)
    (fn c => write c)
  end


(* ---------------------------------------------------------------------- *)
(* Call-by-name *)

(* CBN-conversion:
     - A thunk is a function from unit to a value.
     - The environment maps identifiers to values.
     - gen_cbn maps terms into thunks
     - eval generates a thunk and then forces it

   Roughly...
     #A             =  unit -> A
     #(S->T)        =  unit -> #S -> #T
     #(All(A<S) T)  =  unit -> All(A<S) #T
     #({li:Ti})     =  unit -> {li:#Ti}

     #x             =  x
     #(fun(x:S) e)  =  DELAY(fun(x:#S) #e)
     #(f a)         =  DELAY(FORCE(FORCE(#f) #a))
     #(Fun(A<S) e)  =  DELAY(Fun(A<S) #e)
     #(f [T])       =  DELAY(FORCE(FORCE(f) [T]))
     #({li=ei})     =  DELAY({li=DELAY(ei)})
     #(e.l)         =  DELAY(FORCE(FORCE(e).l))
*)


val TRACE = createflag ["trace"] false;
val trace_count = ref(0)

fun trace s = 
  if !(TRACE)
    then (trace_count := !trace_count + 1;
          implode ["(pr \"[", makestring (!trace_count), "]\"; ", s, ")"])
    else s

fun delay s = "(fn()=> " ^ s ^ ")"

fun cast_app f a = "(app " ^ f ^ " " ^ a ^ ")"

fun force s = cast_app s "()"

fun emit_cbn s = 
    (delay (trace (mapfold (fn s=>s) (fn s1 => fn s2 => s1^s2) "" s)))

fun gen_cbn te env exp = let
  fun ev (VAR(i)) = 
      case_binding env i 
        (fn e' => emit_cbn [e'])
        (fn () => Id.tostr (TEnv.lookup_name te i))
    | ev (ABS(n,t,e1)) = 
        emit_cbn ["fn ", Id.tostr n, " => ", 
              gen_cbn (TEnv.push_binding te n t) (push_env_const env n) e1]
    | ev (APP(e1,e2)) = 
        emit_cbn [force(cast_app 
                          (force (gen_cbn te env e1)) (gen_cbn te env e2))]
    | ev (TABS(x,t,e1)) = 
        gen_cbn (TEnv.push_bound te x t) (push_env_null env x) e1
    | ev (TAPP(e1,t2)) = 
        gen_cbn te env e1
    | ev (RCDV(fields)) = let
        val f = mapappend_tuple 
                  (fn(l,e) => ["(\"", Id.tostr l, "\", cast ", 
                               gen_cbn te env e, ")"])
                  ", " 
                  fields
        in emit_cbn (["["] @ f @ ["]"])
        end
    | ev (PROJ (e,l)) =
        emit_cbn [force (implode
                         ["(assoc (cast ", force (gen_cbn te env e), ")", 
                          " \"", Id.tostr l, "\")"])]
    | ev (PACK(e,_,_)) = gen_cbn te env e
    | ev (OPEN(a,x,e,b)) = let
        in emit_cbn ["let val ", Id.tostr x, " = ", gen_cbn te env e, 
                 " in ", 
                 force (gen_cbn (TEnv.push_binding 
                             (TEnv.push_bound te a Typ.dummytype) 
                             x Typ.dummytype)
                         (push_env_const 
                             (push_env_null env a) x)
                         b),
                 " end"]
        end
    | ev (ASSERT(e,t)) = 
        gen_cbn te env e
    | ev (ABBREV(a,t,e)) = 
        gen_cbn te env (subst_top_type t e) 
    | ev (LET(x,e,b)) = 
        emit_cbn ["let val ", Id.tostr x, " = ", gen_cbn te env e, 
               " in ", 
               force (gen_cbn (TEnv.push_binding te x Typ.dummytype)
                       (push_env_const env x)
                       b),
               " end"]
    | ev (PRIM(p,t)) = 
        emit_cbn [p]
  in wrap DEBUGMUCHO "gen_cbn"
    (fn () => ev exp)
    (fn () => prt te exp)
    (fn c => write c)
  end

val dummycode = "(fn() => raise DummyCode)"

(* ---------------------------------------------------------------------- *)

datatype printercontext = 
    NOPRINTERS 
  | DEEPEN of int * printercontext
  | PRINTER of Typ.T * code * printercontext 

val empty_printercontext = NOPRINTERS

exception BadPrinter

fun define_printer pctx t c = let
  in
    case t of
        Typ.ARROW(tdom,_) => PRINTER(tdom,c,pctx)
      | _ => raise BadPrinter
  end

fun deepen_printers_by (NOPRINTERS) m         = NOPRINTERS
  | deepen_printers_by (DEEPEN(n,pctx)) m     = DEEPEN(n+m,pctx)
  | deepen_printers_by (pctx as PRINTER(_)) m = DEEPEN(m,pctx)

fun deepen_printers pctx = deepen_printers_by pctx 1

val PRINTERS = createflag ["printers"] false
val PRINT = createflag ["print_results"] true

fun eval_print te pctx print_t name = let
  fun p NOPRINTERS _ = NONE
    | p (DEEPEN(n,pctx)) m = p pctx (n+m)
    | p (PRINTER(t,c,pctx)) m = let val t_rel = Typ.relocate m t
                                in if Synth.leq te print_t t_rel
                                   then SOME(c)
                                   else p pctx m
                                end
  in if not (!PRINT) then () else
     case (p pctx 0) of
      NONE => (if !PRINTERS 
                 then write "No printer found\n"
                 else ())
    | SOME(c) => (if !PRINTERS 
                    then (write "Printer: "; write c; write "\n")
                    else ();
                  doit ("val _ = (cast (" ^ c ^ ")) (fn()=>" 
                        ^ nameprefix ^ (Id.tostr name) ^ ") ();\n"))
  end


(* ---------------------------------------------------------------------- *)

fun eval te env name exp = let
  val c = gen_cbn te env exp
  val n = nameprefix ^ (Id.tostr name)
  val sml = "val " ^ n ^ " = " ^ c ^ " ();\n";
  in doit sml;
     n
  end

end
