functor TEnv(
	    structure Globals: GLOBALS
	    structure Typ: TYP
	    sharing Typ.Globals = Globals
	    ) : TENV
	    = struct

structure Globals = Globals
structure Typ = Typ
open Globals
open Pp

datatype BindingType = BND | ABB | VBND  

type T = (Id.T * BindingType * Typ.T) list

(* ------------------------------------------------------------------------ *)

exception UnknownId of string

fun domsize te = List.length te

exception TriedToPopEmptyTEnv
fun pop (bnd::te) = te
  | pop [] = raise TriedToPopEmptyTEnv

fun popn te 0 = te
  | popn te n = popn (pop te) (n-1) 

fun index bvs i =
  let fun ind [] n =
            raise UnknownId(Id.tostr i)
        | ind ((i',_,_)::rest) n =
            if Id.== i i'
               then n
               else ind rest (n+1)
  in ind bvs 0
  end

exception TypeVariableOutOfRange of int

fun lookup_name te i = 
  let val (name,_,_) = (nth (te, i)) in name end

fun bare_lookup te n fbnd fvbnd fabb =
  let val (_,btyp,t) = (nth (te, n)) in 
  case btyp of
     BND => fbnd t
   | VBND => fvbnd t
   | ABB => fabb t
  end

fun lookup te n fbnd fvbnd fabb =
  let val (_,btyp,t) = (nth (te, n)) in 
  case btyp of
     BND => fbnd (Typ.relocate (n+1) t)
   | VBND => fvbnd (Typ.relocate (n+1) t)
   | ABB => fabb (Typ.relocate (n+1) t)
  end
(* ----------------------------------------------------------------------- *)

fun push_bound te i t = (i,BND,t)::te
fun push_abbrev te i t = (i,ABB,t)::te
fun push_binding te i t = (i,VBND,t)::te
val empty = []

(* ----------------------------------------------------------------------- *)

exception WrongKindOfId of T * int * string

fun lookup_bound te i = 
  lookup te i 
    (fn b => b)
    (fn _ => raise WrongKindOfId(te,i,"tvar (not var)"))
    (fn _ => raise WrongKindOfId(te,i,"tvar (not abbrev)"))

fun lookup_bound_or_abbrev te i fb fa = 
  lookup te i 
    fb
    (fn _ => raise WrongKindOfId(te,i,"tvar or abbrev (not var)"))
    fa

fun lookup_binding te i = 
  lookup te i 
    (fn _ => raise WrongKindOfId(te,i,"var (not tvar)"))
    (fn b => b)
    (fn _ => raise WrongKindOfId(te,i,"var (not abbrev)"))

end

